@@ -1,10 +1,10 @@
 #!/usr/bin/perl
 # Parse::DNS::Zone - DNS Zone File Parser
 #
-# Copyright (c) 2009-2011 - Olof Johansson <olof@cpan.org>. 
+# Copyright (c) 2009-2011, 2013, 2015 - Olof Johansson <olof@cpan.org>
 # All rights reserved.
 #
-# This program is free software; you can redistribute it and/or 
+# This program is free software; you can redistribute it and/or
 # modify it under the same terms as Perl itself.
 
 =pod
@@ -34,12 +34,12 @@ Parse::DNS::Zone - DNS Zone File Parser
 =head1 DESCRIPTION
 
 B<Parse::DNS::Zone> parses a zonefile, used to define a DNS Zone
-and gives you the information therein, via an object oriented 
-interface. Parse::DNS::Zone doesn't validate rrdata, except for 
-SOA, and is used to 1) validate the basic structure of the file 
+and gives you the information therein, via an object oriented
+interface. Parse::DNS::Zone doesn't validate rrdata, except for
+SOA, and is used to 1) validate the basic structure of the file
 and 2) extract rdata so you can parse it and validate it yourself.
 
-Parse::DNS::Zone supports the zone file format as described in 
+Parse::DNS::Zone supports the zone file format as described in
 RFC 1034:
 
 =over 4
@@ -52,14 +52,16 @@ RFC 1034:
 
 =back
 
-Parse::DNS::Zone does not support $GENERATE in this version. 
+Parse::DNS::Zone does not support $GENERATE in this version.
 
-=cut 
+=cut
 
 package Parse::DNS::Zone;
-our $VERSION = '0.42';
+our $VERSION = '0.5';
 use warnings;
 use strict;
+use File::Basename;
+use File::Spec;
 use Carp;
 
 =head1 CONSTRUCTOR
@@ -70,16 +72,26 @@ use Carp;
 
 =item B<Required Arguments>
 
-=over 4 
+=over 4
 
 =item * B<origin>
 
 Origin
 
+=back
+
+And additionally, exactly one of the following:
+
+=over 4
+
 =item * B<zonefile>
 
 Path to the zonefile being parsed
 
+=item * B<zonestr>
+
+The zone, as a string.
+
 =back
 
 =item B<Optional Arguments>
@@ -88,18 +100,26 @@ Path to the zonefile being parsed
 
 =item * B<require_soa>
 
-If set to a true value, the parser will whine and die if 
+If set to a true value, the parser will whine and die if
 the zonefile doesn't contain a SOA record. (Default: yes)
 
+=item * B<basepath>
+
+Specify a basepath, from which included relative zonefiles
+should be available. If used with the B<zonefile> parameter,
+this defaults to the directory in which the zonefile is in.
+For $INCLUDEs to work when passing the zone in as a string,
+this needs to be specified.
+
 =item * B<append_origin>
 
 If set to a true value, the parser will append the origin
 to all unqualified domain names (in certain record types,
-currently: CNAME, MX, NS, AFSDB, PTR). If some record 
+currently: CNAME, MX, NS, AFSDB, PTR). If some record
 types are missing from this list, please report that as a
 bug. (Default: no)
 
-This feature do run the risk of becoming stale if new 
+This feature do run the risk of becoming stale if new
 record types are introduced. But if you run into problems,
 don't hesitate to report it!
 
@@ -112,18 +132,31 @@ don't hesitate to report it!
 sub new {
 	my $class = shift;
 	my $self = {
-		require_soa=>1,
-		append_origin=>0,
+		require_soa => 1,
+		append_origin => 0,
 		@_
 	};
 
+	if (not defined $self->{zonestr} and defined $self->{zonefile}) {
+		$self->{zonestr} = _load_zonefile($self->{zonefile});
+	}
+	if (not defined $self->{zonestr}) {
+		croak("You need to specify either zonestr or zonefile");
+	}
+
+	# default basepath is dirname($zonefile)
+	if (not exists $self->{basepath}) {
+		$self->{basepath} = dirname($self->{zonefile}) if
+			defined $self->{zonefile};
+	}
+
 	# append trailing .
 	$self->{origin} .= '.' if($self->{origin}=~/[^[^\.]$/);
 	bless($self, $class);
 
 	_parse($self);
 
-	if($self->{require_soa} && 
+	if($self->{require_soa} &&
 	   (!exists $self->{zone}{$self->{origin}}{soa})) {
 		croak("No SOA in zonefile");
 	}
@@ -139,12 +172,12 @@ sub new {
 
 =head3 $pdz->get_rdata(name=>$name, rr=>$rr, n=>$n, field=>$field)
 
-Is used to get the data associated with a specific name and rr 
-type. The $name can be as the name appears in the zonefile, or a 
-fqdn (with trailing .) as long as it is tracked by the zonefile. 
-If the n argument is specified, the n:th RR in the RRset is 
-returned. Otherwise, you'll get a complete list of the RRset if 
-you're in list context, or the first RR if you're in scalar 
+Is used to get the data associated with a specific name and rr
+type. The $name can be as the name appears in the zonefile, or a
+fqdn (with trailing .) as long as it is tracked by the zonefile.
+If the n argument is specified, the n:th RR in the RRset is
+returned. Otherwise, you'll get a complete list of the RRset if
+you're in list context, or the first RR if you're in scalar
 context.
 
 The $field is the particular component of the resource record to
@@ -166,7 +199,7 @@ sub get_rdata {
 	$name=~s/\.\@\./\.$self->{origin}/g;
 	$name=~s/\.\@$/\.$self->{origin}/g;
 	$name=~s/\@\.$/\.$self->{origin}/g;
-	$name .= ".$self->{origin}" if(($name ne $self->{origin}) && 
+	$name .= ".$self->{origin}" if(($name ne $self->{origin}) &&
 	                               (!($name=~/\.$/)));
 
 	return $self->{zone}{lc $name}{lc $rr}{lc $field}[$n] if defined $n;
@@ -188,7 +221,7 @@ sub exists {
 	$name=~s/\.\@\./\.$self->{origin}/g;
 	$name=~s/\.\@$/\.$self->{origin}/g;
 	$name=~s/\@\.$/\.$self->{origin}/g;
-	$name .= ".$self->{origin}" if(($name ne $self->{origin}) && 
+	$name .= ".$self->{origin}" if(($name ne $self->{origin}) &&
 	                               (!($name=~/\.$/)));
 
 	return exists $self->{zone}{lc $name};
@@ -209,7 +242,7 @@ sub get_rrs {
 	$name=~s/\.\@\./\.$self->{origin}/g;
 	$name=~s/\.\@$/\.$self->{origin}/g;
 	$name=~s/\@\.$/\.$self->{origin}/g;
-	$name .= ".$self->{origin}" if(($name ne $self->{origin}) && 
+	$name .= ".$self->{origin}" if(($name ne $self->{origin}) &&
 	                               (!($name=~/\.$/)));
 
 	foreach my $k (keys %{$self->{zone}{lc $name}}) {
@@ -223,7 +256,7 @@ sub get_rrs {
 
 Returns how many RRs of a given type is defined for $name. For a simple
 setup with a single RR for $name, this will return 1. If you have some
-kind of load balancing or other scheme using multiple RRs of the same 
+kind of load balancing or other scheme using multiple RRs of the same
 type this sub will return the number of "dupes".
 
 =cut
@@ -241,7 +274,7 @@ sub get_dupes {
 	$name=~s/\.\@\./\.$self->{origin}/g;
 	$name=~s/\.\@$/\.$self->{origin}/g;
 	$name=~s/\@\.$/\.$self->{origin}/g;
-	$name .= ".$self->{origin}" if(($name ne $self->{origin}) && 
+	$name .= ".$self->{origin}" if(($name ne $self->{origin}) &&
 	                               (!($name=~/\.$/)));
 
 	return int(@{$self->{zone}{lc $name}{lc $rr}{rdata}});
@@ -279,8 +312,8 @@ sub get_mname {
 
 =head3 $pdz->get_rname( parse=>{0,1} )
 
-Return the RNAME part of the SOA. If parse is set to a value 
-other than 0, the value will be interpreted to show an 
+Return the RNAME part of the SOA. If parse is set to a value
+other than 0, the value will be interpreted to show an
 emailaddress. (default: 0)
 
 =cut
@@ -356,27 +389,45 @@ sub get_minimum {
 	return $self->{soa}{minimum};
 }
 
-1;
-
-# Is used to populate the zone hash used internally. 
+# Is used to populate the zone hash used internally.
 sub _parse {
 	my $self = shift;
-	my %zone = _parse_zone(
-		$self->{zonefile}, $self->{origin}, $self->{append_origin}
+
+	my %zone = $self->_parse_zone(
+		zonestr => $self->{zonestr},
+		origin => $self->{origin},
 	);
 
 	undef $self->{zone};
 	$self->{zone}={%zone};
 }
 
+sub _load_zonefile {
+	my $file = shift;
+	open(my $zonefh, $file) or croak("Could not open $file: $!");
+	return do { local $/; <$zonefh> }; # slurp
+}
+
 # Is used internally to parse a zone from a filename. will do some
 # recursion for the $include, so a procedural implementation is needed
 sub _parse_zone {
+	my $self = shift;
 	# $def_class and $def_ttl are only given when called for included zones
-	my ($zonefile, $origin, $originappend, $def_class, $def_ttl) = @_;
+	my %opts = @_;
 
-	my($zonepath) = $zonefile =~ /^(.*\/)/;
-	open(my $zonefh, $zonefile) or croak("Could not open $zonefile: $!");
+	my $origin = $opts{origin} // $self->{origin};
+
+	my $zonestr = $opts{zonestr};
+	if (not defined $zonestr and exists $opts{zonefile}) {
+		$zonestr = _load_zonefile($opts{zonefile});
+	}
+
+	my ($def_class, $def_ttl);
+	if ($opts{included}) {
+		($def_class, $def_ttl) = @{\%opts}{qw(default_class default_ttl)};
+
+	}
+	my $zonepath = $self->{basepath};
 
 	my $mrow;
 	my $prev;
@@ -393,13 +444,13 @@ sub _parse_zone {
 		(\S+)\s+ # type
 		(.*) # rdata
 	$/ix;
-	
-	while(<$zonefh>) {
+
+	for (split /\n/, $zonestr) {
 		chomp;
 		s/;.*$//;
 		next if /^\s*$/;
 		s/\s+/ /g;
-		
+
 		s/^\@ /$origin /g;
 		s/ \@ / $origin /g;
 		s/ \@$/ $origin/g;
@@ -407,14 +458,14 @@ sub _parse_zone {
 		# handles mutlirow entries, with ()
 		if($mrow) {
 			$mrow.=$_;
-			
-			next if(! /\)/); 
 
-			# End of multirow 
+			next if(! /\)/);
+
+			# End of multirow
 			$mrow=~s/[\(\)]//g;
 			$mrow=~s/\n//mg;
 			$mrow=~s/\s+/ /g;
-			$mrow .= "\n";	
+			$mrow .= "\n";
 
 			$_ = $mrow;
 			undef $mrow;
@@ -435,14 +486,16 @@ sub _parse_zone {
 
 			my $zfile = $1;
 			if($1 !~ m/^\//) {
-				$zfile = $zonepath.$zfile;
+				$zfile = File::Spec->catfile($zonepath, $zfile);
 			}
 
-			my
-				%subz=_parse_zone(
-					$zfile, $subo, $originappend, 
-					$def_class, $def_ttl
-				);
+			my %subz = $self->_parse_zone(
+				zonefile => $zfile,
+				included => 1,
+				origin => $subo,
+				default_class => $def_class,
+				default_ttl => $def_ttl,
+			);
 
 			foreach my $k (keys %subz) {
 				$zone{$k}=$subz{$k};
@@ -483,9 +536,10 @@ sub _parse_zone {
 		}
 
 		$prev=$name;
-		$name.=".$origin" if $name ne $origin && $name !~ /\.$/;
+		$name = _fqdnize($name, $origin);
 
-		if($originappend and $type =~ /^(?:cname|afsdb|mx|ns)$/i and 
+		if($self->{append_origin} and
+		   $type =~ /^(?:cname|afsdb|mx|ns)$/i and
 		   $rdata ne $origin and $rdata !~ /\.$/) {
 			$rdata.=".$origin";
 		}
@@ -495,18 +549,28 @@ sub _parse_zone {
 		push(@{$zone{lc $name}{lc $type}{class}}, $class);
 	}
 
-	close $zonefh;
 	return %zone;
 }
 
-# Is used to parse the SOA and build the soa hash as used 
+sub _fqdnize {
+	my ($name, $origin) = @_;
+
+	$origin //= '.';
+	$origin .= '.' unless $origin =~ /\.$/;
+
+	return $name if $name =~ /\.$/;
+	return "$name." if $origin eq '.';
+	return "$name.$origin";
+}
+
+# Is used to parse the SOA and build the soa hash as used
 # internally..
 sub _parse_soa {
 	my $self = shift;
 	my $soa_rd = get_rdata($self, (name=>"$self->{origin}", rr=>'SOA'));
 	my($mname,$rname,$serial,$refresh,$retry,$expire,$minimum)=
 		$soa_rd=~/^(\S+) (\S+) (\d+) (\d+) (\d+) (\d+) (\d+)\s*$/;
-	
+
 	$self->{soa}{mname}=$mname;
 	$self->{soa}{rname}=$rname;
 	$self->{soa}{serial}=$serial;
@@ -516,6 +580,8 @@ sub _parse_soa {
 	$self->{soa}{minimum}=$minimum;
 }
 
+1;
+
 =head1 SEE ALSO
 
 RFC 1034, RFC 1035, Bind Administrator's Guide
@@ -527,10 +593,11 @@ version is available on https://github.com/olof/Parse-DNS-Zone.
 
 =head1 COPYRIGHT
 
-Copyright (c) 2009-2011 - Olof Johansson <olof@cpan.org>. 
+ Copyright (c) 2009-2011, 2013, 2015 - Olof Johansson <olof@cpan.org>
+
 All rights reserved.
 
-This program is free software; you can redistribute it and/or 
+This program is free software; you can redistribute it and/or
 modify it under the same terms as Perl itself.
 
 =cut
